@@ -31,8 +31,10 @@
  *   <http://technet.microsoft.com/en-us/library/ee323524%28office.13%29.aspx>
  *  Microsoft Office Communications Server 2007 R2 Technical Reference Guide
  *   <http://go.microsoft.com/fwlink/?LinkID=159649>
- *  XML XCCOS message specification
- *   <???> (searches on the internet currently reveal nothing)
+ *  Microsoft DevNet: [MS-XCCOSIP] Extensible Chat Control Over SIP
+ *   <http://msdn.microsoft.com/en-us/library/hh624112.aspx>
+ *  RFC 4028: Session Timers in the Session Initiation Protocol (SIP)
+ *   <http://www.rfc-editor.org/rfc/rfc4028.txt>
  *
  *
  * @TODO:
@@ -73,15 +75,6 @@
  *
  *     use this to sync chats in buddy list on multiple clients?
  *
- *   - cmd:bccontext
- *     send after cmd:join to trigger rpl:bccontext
- *
- *   - rpl:bccontext
- *     according to available documentation delivers channel history, etc.
- *     [no log file examples]
- *     can we add the history to the window when we open the window, or would
- *     that confuse users that use history based on client log?
- *
  *   - cmd:getinv
  *       <inv inviteId="1" domain="<DOMAIN>" />
  *     rpl:getinv
@@ -92,7 +85,7 @@
  *     should we automatically join those channels or ask user to join/add?
  *
  *   - chatserver_command_message()
- *     needs to support multiple <chatgrp> nodes?
+ *     needs to support multiple <grpchat> nodes?
  *     [no log file examples]
  *
  *   - create/delete chat rooms
@@ -163,6 +156,7 @@ struct sipe_groupchat {
 	GHashTable *uri_to_chat_session;
 	GHashTable *msgs;
 	guint envid;
+	guint expires;
 	gboolean connected;
 };
 
@@ -246,9 +240,12 @@ void sipe_groupchat_init(struct sipe_core_private *sipe_private)
 {
 	const gchar *setting = sipe_backend_setting(SIPE_CORE_PUBLIC,
 						    SIPE_SETTING_GROUPCHAT_USER);
-	gboolean user_set = !is_empty(setting);
-	gchar **parts = g_strsplit(user_set ? setting : sipe_private->username,
-				   "@", 2);
+	const gchar *persistent = sipe_private->persistentChatPool_uri;
+	gboolean user_set    = !is_empty(setting);
+	gboolean provisioned = !is_empty(persistent);
+	gchar **parts = g_strsplit(user_set ? setting :
+				   provisioned ? persistent :
+				   sipe_private->username, "@", 2);
 	gboolean domain_found = !is_empty(parts[1]);
 	const gchar *user = "ocschat";
 	const gchar *domain = parts[domain_found ? 1 : 0];
@@ -256,13 +253,14 @@ void sipe_groupchat_init(struct sipe_core_private *sipe_private)
 	struct sip_session *session;
 	struct sipe_groupchat *groupchat;
 
-	/* User specified valid 'user@company.com' */
-	if (user_set && domain_found && !is_empty(parts[0]))
+	/* User specified or provisioned URI is valid 'user@company.com' */
+	if ((user_set || provisioned) && domain_found && !is_empty(parts[0]))
 		user = parts[0];
 
-	SIPE_DEBUG_INFO("sipe_groupchat_init: username '%s' setting '%s' split '%s'/'%s' GC user %s@%s",
-			sipe_private->username, setting ? setting : "(null)", parts[0],
-			parts[1] ? parts[1] : "(null)", user, domain);
+	SIPE_DEBUG_INFO("sipe_groupchat_init: username '%s' setting '%s' persistent '%s' split '%s'/'%s' GC user %s@%s",
+			sipe_private->username, setting ? setting : "(null)",
+			persistent ? persistent : "(null)",
+			parts[0], parts[1] ? parts[1] : "(null)", user, domain);
 
 	if (!sipe_private->groupchat)
 		sipe_groupchat_allocate(sipe_private);
@@ -299,7 +297,7 @@ static void groupchat_init_retry(struct sipe_core_private *sipe_private)
 	groupchat->connected = FALSE;
 
 	sipe_schedule_seconds(sipe_private,
-			      "<+grouchat-retry>",
+			      "<+groupchat-retry>",
 			      NULL,
 			      GROUPCHAT_RETRY_TIMEOUT,
 			      groupchat_init_retry_cb,
@@ -364,11 +362,65 @@ static gchar *generate_chanid_node(const gchar *uri, guint key)
 	return chanid;
 }
 
+/* TransCallback */
+static void groupchat_update_cb(struct sipe_core_private *sipe_private,
+				gpointer data);
+static gboolean groupchat_expired_session_response(struct sipe_core_private *sipe_private,
+						   struct sipmsg *msg,
+						   SIPE_UNUSED_PARAMETER struct transaction *trans)
+{
+	struct sipe_groupchat *groupchat = sipe_private->groupchat;
+
+	/* 481 Call Leg Does Not Exist -> server dropped session */
+	if (msg->response == 481) {
+		struct sip_session *session = groupchat->session;
+		struct sip_dialog *dialog = sipe_dialog_find(session,
+							     session->with);
+
+		/* close dialog from our side */
+		sip_transport_bye(sipe_private, dialog);
+		sipe_dialog_remove(session, session->with);
+		/* dialog is no longer valid */
+
+		/* re-initialize groupchat session */
+		groupchat->session = NULL;
+		groupchat->connected = FALSE;
+		sipe_groupchat_init(sipe_private);
+	} else {
+		sipe_schedule_seconds(sipe_private,
+				      "<+groupchat-expires>",
+				      NULL,
+				      groupchat->expires,
+				      groupchat_update_cb,
+				      NULL);
+	}
+
+	return(TRUE);
+}
+
+/* sipe_schedule_action */
+static void groupchat_update_cb(struct sipe_core_private *sipe_private,
+				SIPE_UNUSED_PARAMETER gpointer data)
+{
+	struct sipe_groupchat *groupchat = sipe_private->groupchat;
+
+	if (groupchat->session) {
+		struct sip_dialog *dialog = sipe_dialog_find(groupchat->session,
+							     groupchat->session->with);
+
+		if (dialog)
+			sip_transport_update(sipe_private,
+					     dialog,
+					     groupchat_expired_session_response);
+	}
+}
+
 static struct sipe_groupchat_msg *chatserver_command(struct sipe_core_private *sipe_private,
 						     const gchar *cmd);
 
 void sipe_groupchat_invite_response(struct sipe_core_private *sipe_private,
-				    struct sip_dialog *dialog)
+				    struct sip_dialog *dialog,
+				    struct sipmsg *response)
 {
 	struct sipe_groupchat *groupchat = sipe_private->groupchat;
 
@@ -378,6 +430,9 @@ void sipe_groupchat_invite_response(struct sipe_core_private *sipe_private,
 		/* response to initial invite */
 		struct sipe_groupchat_msg *msg = generate_xccos_message(groupchat,
 									"<cmd id=\"cmd:requri\" seqid=\"1\"><data/></cmd>");
+		const gchar *session_expires = sipmsg_find_header(response,
+								  "Session-Expires");
+
 		sip_transport_info(sipe_private,
 				   "Content-Type: text/plain\r\n",
 				   msg->xccos,
@@ -385,6 +440,24 @@ void sipe_groupchat_invite_response(struct sipe_core_private *sipe_private,
 				   NULL);
 		sipe_groupchat_msg_remove(msg);
 
+		if (session_expires) {
+			groupchat->expires = strtoul(session_expires, NULL, 10);
+
+			if (groupchat->expires) {
+				SIPE_DEBUG_INFO("sipe_groupchat_invite_response: session expires in %d seconds",
+						groupchat->expires);
+
+				if (groupchat->expires > 10)
+					groupchat->expires -= 10;
+				sipe_schedule_seconds(sipe_private,
+						      "<+groupchat-expires>",
+						      NULL,
+						      groupchat->expires,
+						      groupchat_update_cb,
+						      NULL);
+			}
+		}
+
 	} else {
 		/* response to group chat server invite */
 		gchar *invcmd;
@@ -426,6 +499,22 @@ void sipe_groupchat_invite_response(struct sipe_core_private *sipe_private,
 	}
 }
 
+static void chatserver_command_error_notify(struct sipe_core_private *sipe_private,
+					    struct sipe_chat_session *chat_session,
+					    const gchar *content)
+{
+	gchar *label  = g_strdup_printf(_("This message was not delivered to chat room '%s'"),
+					chat_session->title);
+	gchar *errmsg = g_strdup_printf("%s:\n<font color=\"#888888\"></b>%s<b></font>",
+					label, content);
+	g_free(label);
+	sipe_backend_notify_message_error(SIPE_CORE_PUBLIC,
+					  chat_session->backend,
+					  NULL,
+					  errmsg);
+	g_free(errmsg);
+}
+
 /* TransCallback */
 static gboolean chatserver_command_response(struct sipe_core_private *sipe_private,
 					    struct sipmsg *msg,
@@ -437,18 +526,12 @@ static gboolean chatserver_command_response(struct sipe_core_private *sipe_priva
 
 		SIPE_DEBUG_INFO("chatserver_command_response: failure %d", msg->response);
 
-		if (chat_session) {
-			gchar *label  = g_strdup_printf(_("This message was not delivered to chat room '%s'"),
-							chat_session->title);
-			gchar *errmsg = g_strdup_printf("%s:\n<font color=\"#888888\"></b>%s<b></font>",
-							label, gmsg->content);
-			g_free(label);
-			sipe_backend_notify_message_error(SIPE_CORE_PUBLIC,
-							  chat_session->backend,
-							  NULL,
-							  errmsg);
-			g_free(errmsg);
-		}
+		if (chat_session)
+			chatserver_command_error_notify(sipe_private,
+							chat_session,
+							gmsg->content);
+
+		groupchat_expired_session_response(sipe_private, msg, trans);
 	}
 	return TRUE;
 }
@@ -457,21 +540,25 @@ static struct sipe_groupchat_msg *chatserver_command(struct sipe_core_private *s
 						     const gchar *cmd)
 {
 	struct sipe_groupchat *groupchat = sipe_private->groupchat;
-	struct sipe_groupchat_msg *msg = generate_xccos_message(groupchat, cmd);
-
 	struct sip_dialog *dialog = sipe_dialog_find(groupchat->session,
 						     groupchat->session->with);
+	struct sipe_groupchat_msg *msg = NULL;
 
-	struct transaction_payload *payload = g_new0(struct transaction_payload, 1);
-	struct transaction *trans = sip_transport_info(sipe_private,
-						       "Content-Type: text/plain\r\n",
-						       msg->xccos,
-						       dialog,
-						       chatserver_command_response);
+	if (dialog) {
+		struct transaction_payload *payload = g_new0(struct transaction_payload, 1);
+		struct transaction *trans;
 
-	payload->destroy = sipe_groupchat_msg_remove;
-	payload->data    = msg;
-	trans->payload   = payload;
+		msg = generate_xccos_message(groupchat, cmd);
+		trans = sip_transport_info(sipe_private,
+					   "Content-Type: text/plain\r\n",
+					   msg->xccos,
+					   dialog,
+					   chatserver_command_response);
+
+		payload->destroy = sipe_groupchat_msg_remove;
+		payload->data    = msg;
+		trans->payload   = payload;
+	}
 
 	return(msg);
 }
@@ -499,7 +586,7 @@ static void chatserver_response_uri(struct sipe_core_private *sipe_private,
 			session->is_groupchat = TRUE;
 			sipe_im_invite(sipe_private, session, uri, NULL, NULL, NULL, FALSE);
 		} else {
-			SIPE_DEBUG_WARNING_NOFORMAT("process_incoming_info_groupchat: no server URI found!");
+			SIPE_DEBUG_WARNING_NOFORMAT("chatserver_response_uri: no server URI found!");
 			groupchat_init_retry(sipe_private);
 		}
 }
@@ -648,7 +735,7 @@ static void chatserver_response_join(struct sipe_core_private *sipe_private,
 											     uri);
 				gboolean new = (chat_session == NULL);
 				const gchar *attr = sipe_xml_attribute(node, "name");
-				char *self = sip_uri_self(sipe_private);
+				gchar *self = sip_uri_self(sipe_private);
 				const sipe_xml *aib;
 
 				if (new) {
@@ -709,6 +796,16 @@ static void chatserver_response_join(struct sipe_core_private *sipe_private,
 						g_strfreev(ids);
 					}
 				}
+
+				/* Request last 25 entries from channel history */
+				self = g_strdup_printf("<cmd id=\"cmd:bccontext\" seqid=\"1\">"
+						       "<data>"
+						       "<chanib uri=\"%s\"/>"
+						       "<bcq><last cnt=\"25\"/></bcq>"
+						       "</data>"
+						       "</cmd>", chat_session->id);
+				chatserver_command(sipe_private, self);
+				g_free(self);
 			}
 		}
 
@@ -716,6 +813,25 @@ static void chatserver_response_join(struct sipe_core_private *sipe_private,
 	}
 }
 
+static void chatserver_grpchat_message(struct sipe_core_private *sipe_private,
+				       const sipe_xml *grpchat);
+
+static void chatserver_response_history(SIPE_UNUSED_PARAMETER struct sipe_core_private *sipe_private,
+					SIPE_UNUSED_PARAMETER struct sip_session *session,
+					SIPE_UNUSED_PARAMETER guint result,
+					SIPE_UNUSED_PARAMETER const gchar *message,
+					const sipe_xml *xml)
+{
+	const sipe_xml *grpchat;
+
+	for (grpchat = sipe_xml_child(xml, "chanib/msg");
+	     grpchat;
+	     grpchat = sipe_xml_twin(grpchat))
+		if (sipe_strequal(sipe_xml_attribute(grpchat, "id"),
+				  "grpchat"))
+			chatserver_grpchat_message(sipe_private, grpchat);
+}
+
 static void chatserver_response_part(struct sipe_core_private *sipe_private,
 				     SIPE_UNUSED_PARAMETER struct sip_session *session,
 				     guint result,
@@ -837,14 +953,15 @@ static const struct response {
 			       guint result, const gchar *,
 			       const sipe_xml *xml);
 } response_table[] = {
-	{ "rpl:requri",   chatserver_response_uri },
-	{ "rpl:chansrch", chatserver_response_channel_search },
-	{ "rpl:join",     chatserver_response_join },
-	{ "rpl:bjoin",    chatserver_response_join },
-	{ "rpl:part",     chatserver_response_part },
-	{ "ntc:join",     chatserver_notice_join },
-	{ "ntc:bjoin",    chatserver_notice_join },
-	{ "ntc:part",     chatserver_notice_part },
+	{ "rpl:requri",    chatserver_response_uri },
+	{ "rpl:chansrch",  chatserver_response_channel_search },
+	{ "rpl:join",      chatserver_response_join },
+	{ "rpl:bjoin",     chatserver_response_join },
+	{ "rpl:bccontext", chatserver_response_history },
+	{ "rpl:part",      chatserver_response_part },
+	{ "ntc:join",      chatserver_notice_join },
+	{ "ntc:bjoin",     chatserver_notice_join },
+	{ "ntc:part",      chatserver_notice_part },
 	{ NULL, NULL }
 };
 
@@ -894,12 +1011,13 @@ static void chatserver_response(struct sipe_core_private *sipe_private,
 }
 
 static void chatserver_grpchat_message(struct sipe_core_private *sipe_private,
-				       const sipe_xml *chatgrp)
+				       const sipe_xml *grpchat)
 {
 	struct sipe_groupchat *groupchat = sipe_private->groupchat;
-	const gchar *uri = sipe_xml_attribute(chatgrp, "chanUri");
-	const gchar *from = sipe_xml_attribute(chatgrp, "author");
-	gchar *text = sipe_xml_data(sipe_xml_child(chatgrp, "chat"));
+	const gchar *uri = sipe_xml_attribute(grpchat, "chanUri");
+	const gchar *from = sipe_xml_attribute(grpchat, "author");
+	time_t when = sipe_utils_str_to_time(sipe_xml_attribute(grpchat, "ts"));
+	gchar *text = sipe_xml_data(sipe_xml_child(grpchat, "chat"));
 	struct sipe_chat_session *chat_session;
 	gchar *escaped;
 
@@ -923,7 +1041,7 @@ static void chatserver_grpchat_message(struct sipe_core_private *sipe_private,
 	escaped = g_markup_escape_text(text, -1);
 	g_free(text);
 	sipe_backend_chat_message(SIPE_CORE_PUBLIC, chat_session->backend,
-				  from, escaped);
+				  from, when, escaped);
 	g_free(escaped);
 }
 
@@ -933,19 +1051,40 @@ void process_incoming_info_groupchat(struct sipe_core_private *sipe_private,
 {
 	sipe_xml *xml = sipe_xml_parse(msg->body, msg->bodylen);
 	const sipe_xml *node;
+	const gchar *callid;
+	struct sip_dialog *dialog;
 
-	/* @TODO: is this always correct?*/
-	sip_transport_response(sipe_private, msg, 200, "OK", NULL);
+	callid = sipmsg_find_header(msg, "Call-ID");
+	dialog = sipe_dialog_find(session, session->with);
+	if (sipe_strequal(callid, dialog->callid)) {
 
-	if (!xml) return;
+		sip_transport_response(sipe_private, msg, 200, "OK", NULL);
+
+		if        (((node = sipe_xml_child(xml, "rpl")) != NULL) ||
+			   ((node = sipe_xml_child(xml, "ntc")) != NULL)) {
+			chatserver_response(sipe_private, node, session);
+		} else if ((node = sipe_xml_child(xml, "grpchat")) != NULL) {
+			chatserver_grpchat_message(sipe_private, node);
+		} else {
+			SIPE_DEBUG_INFO_NOFORMAT("process_incoming_info_groupchat: ignoring unknown response");
+		}
 
-	if        (((node = sipe_xml_child(xml, "rpl")) != NULL) ||
-		   ((node = sipe_xml_child(xml, "ntc")) != NULL)) {
-		chatserver_response(sipe_private, node, session);
-	} else if ((node = sipe_xml_child(xml, "grpchat")) != NULL) {
-		chatserver_grpchat_message(sipe_private, node);
 	} else {
-		SIPE_DEBUG_INFO_NOFORMAT("process_incoming_info_groupchat: ignoring unknown response");
+		/*
+		 * Our last session got disconnected without proper shutdown,
+		 * e.g. by Pidgin crashing or network connection loss. When
+		 * we reconnect to the group chat the server will send INFO
+		 * messages to the current *AND* the obsolete Call-ID, until
+		 * the obsolete session expires.
+		 *
+		 * Ignore these INFO messages to avoid, e.g. duplicate texts,
+		 * and respond with an error so that the server knows that we
+		 * consider this dialog to be terminated.
+		 */
+		SIPE_DEBUG_INFO("process_incoming_info_groupchat: ignoring unsolicited INFO message to obsolete Call-ID: %s",
+				callid);
+
+		sip_transport_response(sipe_private, msg, 487, "Request Terminated", NULL);
 	}
 
 	sipe_xml_free(xml);
@@ -957,6 +1096,7 @@ void sipe_groupchat_send(struct sipe_core_private *sipe_private,
 {
 	struct sipe_groupchat *groupchat = sipe_private->groupchat;
 	gchar *cmd, *self, *timestamp, *tmp;
+	gchar **lines, **strvp;
 	struct sipe_groupchat_msg *msg;
 
 	if (!groupchat || !chat_session)
@@ -976,12 +1116,24 @@ void sipe_groupchat_send(struct sipe_core_private *sipe_private,
 	 *    < -> &lt;
 	 *    & -> &amp;
 	 *
-	 * No need to escape them here.
+	 * Group Chat only accepts plain text, not full HTML. So we have to
+	 * strip all HTML tags and XML escape the text.
 	 *
-	 * Only exception are line breaks which are encoded as <br>.
-	 * Replace them with the correct XML tag <br/>.
+	 * Line breaks are encoded as <br> and therefore need to be replaced
+	 * before stripping. In order to prevent HTML stripping to strip line
+	 * endings, we need to split the text into lines on <br>.
 	 */
-        tmp = replace(what, "<br>", "<br/>");
+	lines = g_strsplit(what, "<br>", 0);
+	for (strvp = lines; *strvp; strvp++) {
+		/* replace array entry with HTML stripped & XML escaped version */
+		gchar *stripped = sipe_backend_markup_strip_html(*strvp);
+		gchar *escaped  = g_markup_escape_text(stripped, -1);
+		g_free(stripped);
+		g_free(*strvp);
+		*strvp = escaped;
+	}
+	tmp = g_strjoinv("\r\n", lines);
+	g_strfreev(lines);
 	cmd = g_strdup_printf("<grpchat id=\"grpchat\" seqid=\"1\" chanUri=\"%s\" author=\"%s\" ts=\"%s\">"
 			      "<chat>%s</chat>"
 			      "</grpchat>",
@@ -992,8 +1144,14 @@ void sipe_groupchat_send(struct sipe_core_private *sipe_private,
 	msg = chatserver_command(sipe_private, cmd);
 	g_free(cmd);
 
-	msg->session = chat_session;
-	msg->content = g_strdup(what);
+	if (msg) {
+		msg->session = chat_session;
+		msg->content = g_strdup(what);
+	} else {
+		chatserver_command_error_notify(sipe_private,
+						chat_session,
+						what);
+	}
 }
 
 void sipe_groupchat_leave(struct sipe_core_private *sipe_private,
