@@ -426,7 +426,11 @@ static void password_ok_cb(PurpleConnection *gc,
 
 		if (purple_request_fields_get_bool(fields, "remember"))
 			purple_account_set_remember_password(account, TRUE);
-		purple_account_set_password(account, password);
+		purple_account_set_password(account, password
+#if PURPLE_VERSION_CHECK(3,0,0)
+					    , NULL, NULL
+#endif
+					   );
 
 		/* Now we have a password and we can connect */
 		connect_to_core(gc, account, password);
@@ -486,19 +490,35 @@ static int sipe_purple_send_im(PurpleConnection *gc,
 	return 1;
 }
 
-#define SIPE_TYPING_SEND_TIMEOUT 4
-
 static unsigned int sipe_purple_send_typing(PurpleConnection *gc,
 					    const char *who,
 					    PurpleTypingState state)
 {
-	if (state == PURPLE_NOT_TYPING)
-		return 0;
+	gboolean typing = (state == PURPLE_TYPING);
 
-	sipe_core_user_feedback_typing(PURPLE_GC_TO_SIPE_CORE_PUBLIC,
-				       who);
+	/* only enable this debug output while testing
+	   SIPE_DEBUG_INFO("sipe_purple_send_typing: '%s' state %d", who, state); */
 
-	return SIPE_TYPING_SEND_TIMEOUT;
+	/*
+	 * libpurple calls this function with PURPLE_NOT_TYPING *after*
+	 * calling sipe_purple_send_im() with the message. This causes
+	 * SIPE core to send out two SIP messages to the same dialog in
+	 * short succession without waiting for the response to the first
+	 * one. Some servers then reject the first one with
+	 *
+	 *    SIP/2.0 500 Stale CSeq Value
+	 *
+	 * which triggers a "message not delivered" error for the user.
+	 *
+	 * Work around this by filtering out PURPLE_NOT_TYPING events.
+	 */
+	if (state != PURPLE_NOT_TYPING)
+		sipe_core_user_feedback_typing(PURPLE_GC_TO_SIPE_CORE_PUBLIC,
+					       who,
+					       typing);
+
+	/* tell libpurple to send typing indications every 4 seconds */
+	return(typing ? 4 : 0);
 }
 
 static void sipe_purple_get_info(PurpleConnection *gc, const char *who)
